#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implementation of metamodel as either PC, aPC or GPE
"""

import copy
import os
import warnings

import matplotlib.pyplot as plt
import numpy as np
import sklearn.linear_model as lm
from joblib import Parallel, delayed
from scipy.optimize import minimize, NonlinearConstraint
from sklearn.decomposition import PCA as sklearnPCA
from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm

from .apoly_construction import apoly_construction
from .bayes_linear import VBLinearRegression, EBLinearRegression
from .eval_rec_rule import eval_univ_basis
from .glexindex import glexindex
from .input_space import InputSpace
from .orthogonal_matching_pursuit import OrthogonalMatchingPursuit
from .reg_fast_ard import RegressionFastARD
from .reg_fast_laplace import RegressionFastLaplace

warnings.filterwarnings("ignore")
# Load the mplstyle
# noinspection SpellCheckingInspection
plt.style.use(os.path.join(os.path.split(__file__)[0],
                           '../', 'bayesvalidrox.mplstyle'))


class MetaModel:
    """
    Meta (surrogate) model base class

    This class describes the necessary functions and propoerties of a 
    surrogate model in bayesvalidrox. It accepts an input object (input_obj)
    containing the specification of the distributions for uncertain parameters.

    Attributes
    ----------
    # TODO: recheck these descriptions
    input_obj : obj
        Input object with the information on the model input parameters.
    bootstrap_method : str
        Bootstraping method. Options are `'normal'` and `'fast'`. The default
        is `'fast'`. It means that in each iteration except the first one, only
        the coefficent are recalculated with the ordinary least square method.
    n_bootstrap_itrs : int
        Number of iterations for the bootstrap sampling. The default is `1`.
    dim_red_method : str
        Dimensionality reduction method for the output space. The available
        method is based on principal component analysis (PCA). The Default is
        `'no'`. There are two ways to select number of components: use
        percentage of the explainable variance threshold (between 0 and 100)
        (Option A) or direct prescription of components' number (Option B):
            >>> MetaModelOpts = MetaModel()
            >>> MetaModelOpts.dim_red_method = 'PCA'
            >>> MetaModelOpts.var_pca_threshold = 99.999  # Option A
            >>> MetaModelOpts.n_pca_components = 12 # Option B
    is_gaussian : bool
        Set to True if the surrogate model returns mean and stdev.
        The default is `False`.
    verbose : bool
        Prints summary of the regression results. Default is `False`.
        
    """

    def __init__(self, input_obj, bootstrap_method='fast',
                 n_bootstrap_itrs=1, dim_red_method='no', is_gaussian = False, 
                 verbose=False):

        self.input_obj = input_obj
        self.bootstrap_method = bootstrap_method
        self.n_bootstrap_itrs = n_bootstrap_itrs
        self.dim_red_method = dim_red_method
        self.verbose = verbose
        
        self.is_gaussian = is_gaussian

        # Other params
        self.InputSpace = None
        self.var_pca_threshold = None
        self.errorScale = None
        self.errorclf_poly = None
        self.errorRegMethod = None
        self.nlc = None
        self.n_pca_components = None
        self.out_names = None
        self.n_samples = None
        self.CollocationPoints = None
        self.pca = None
        self.LCerror = None
        self.score_dict = None
        self.basis_dict = None
        self.x_scaler = None
        self.gp_poly = None
        self.n_params = None
        self.ndim = None
        self.init_type = None
        self.rmse = None
        
    def check_is_gaussian(self) -> bool:
        """
        """
        None

    def build_metamodel(self, n_init_samples=None) -> None:
        """
        Builds the parts for the metamodel (polynomes,...) that are neede before fitting.

        Returns
        -------
        None
            DESCRIPTION.

        """
        None

    def fit(self, X: np.array, y: dict, parallel=False, verbose=False):
        """
        Fits the surrogate to the given data (samples X, outputs y).
        Note here that the samples X should be the transformed samples provided
        by the experimental design if the transformation is used there.

        Parameters
        ----------
        X : 2D list or np.array of shape (#samples, #dim)
            The parameter value combinations that the model was evaluated at.
        y : dict of 2D lists or arrays of shape (#samples, #timesteps)
            The respective model evaluations.
        parallel : bool
            Set to True to run the training in parallel for various keys.
            The default is False.
        verbose : bool
            Set to True to obtain more information during runtime.
            The default is False.

        Returns
        -------
        None.

        """
        None

    # -------------------------------------------------------------------------
    def pca_transformation(self, target):
        """
        Transforms the targets (outputs) via Principal Component Analysis.
        The number of features is set by `self.n_pca_components`.
        If this is not given, `self.var_pca_threshold` is used as a threshold.

        Parameters
        ----------
        target : array of shape (n_samples,)
            Target values.

        Returns
        -------
        pca : obj
            Fitted sklearnPCA object.
        OutputMatrix : array of shape (n_samples,)
            Transformed target values.
        n_pca_components : int
            Number of selected principal components.

        """
        # Get current shape of the outputs
        n_samples, n_features = target.shape
        
        # Use the given n_pca_components
        n_pca_components = self.n_pca_components
        
        # Switch to var_pca if n_pca_components is too large
        if (n_pca_components is not None) and (n_pca_components > n_features):
            n_pca_components = None
            if self.verbose:
                print('')
                print('WARNING: Too many components are set for PCA. The transformation will proceed based on explainable variance.')
        
        # Calculate n_pca_components based on decomposition of the variance
        if n_pca_components is None:
            if self.var_pca_threshold is not None:
                var_pca_threshold = self.var_pca_threshold
            else:
                var_pca_threshold = 99#100.0
            # Instantiate and fit sklearnPCA object
            covar_matrix = sklearnPCA(n_components=None)
            covar_matrix.fit(target)
            var = np.cumsum(np.round(covar_matrix.explained_variance_ratio_,
                                     decimals=5) * 100)
            # Find the number of components to explain self.varPCAThreshold of
            # variance
            try:
                n_components = np.where(var >= var_pca_threshold)[0][0] + 1
            except IndexError:
                n_components = min(n_samples, n_features)

            n_pca_components = min(n_samples, n_features, n_components)

        # Print out a report
        #if self.verbose:
        #    print()
        #    print('-' * 50)
        #    print(f"PCA transformation is performed with {n_pca_components}"
        #          " components.")
        #    print('-' * 50)
        #    print()

        # Set the solver to 'auto' if no reduction is wanted
        # Otherwise use 'arpack'
        solver = 'auto'
        if n_pca_components < n_features:
            solver = 'arpack'
            
        # Fit and transform with the selected number of components
        pca = sklearnPCA(n_components=n_pca_components, svd_solver=solver)
        scaled_target = pca.fit_transform(target)

        return pca, scaled_target, n_pca_components

    # -------------------------------------------------------------------------
    def eval_metamodel(self, samples):
        """
        Evaluates metamodel at the requested samples. One can also generate
        nsamples.

        Parameters
        ----------
        samples : array of shape (n_samples, n_params), optional
            Samples to evaluate metamodel at. The default is None.

        Returns
        -------
        mean_pred : dict
            Mean of the predictions.
        std_pred : dict
            Standard deviatioon of the predictions. Return None if 
            self.is_gaussian == False
        """
        return None, None

    # -------------------------------------------------------------------------
    class auto_vivification(dict):
        """
        Implementation of perl's AutoVivification feature.

        Source: https://stackoverflow.com/a/651879/18082457
        """

        def __getitem__(self, item):
            try:
                return dict.__getitem__(self, item)
            except KeyError:
                value = self[item] = type(self)()
                return value

    # -------------------------------------------------------------------------
    def copy_meta_model_opts(self):
        """
        This method is a convinient function to copy the metamodel options.

        Returns
        -------
        new_MetaModelOpts : object
            The copied object.

        """
        # TODO: what properties should be moved to the new object?
        new_MetaModelOpts = copy.deepcopy(self)
        new_MetaModelOpts.input_obj = self.input_obj  # InputObj
        new_MetaModelOpts.InputSpace = self.InputSpace
        new_MetaModelOpts.n_params = len(self.input_obj.Marginals)
        return new_MetaModelOpts

    # -------------------------------------------------------------------------
    def add_InputSpace(self):
        """
        Instanciates experimental design object.

        Returns
        -------
        None.

        """
        self.InputSpace = InputSpace(self.input_obj)
